#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mach.h>
#include "../cmd/nc/n.out.h"

/*
 * Nios-specific debugger interface
 */

typedef struct	Instr	Instr;
struct	Instr
{
	Map	*map;
	ulong	w;
	uvlong	addr;
	int		op;
	char	ra, rb, rc;
	ulong	imm26;
	short	imm16;
	short	imm5;

	char*	curr;			/* fill point in buffer */
	char*	end;			/* end of buffer */
};

typedef struct	Opcode	Opcode;
struct	Opcode
{
	char 	as;				/* symbolic opcode (from n.out.h) */
	char	fmt;			/* index into fmts table */
};

/* copy anames from compiler */
static
#include "../cmd/nc/enam.c"

static char *fmts[] = {
 /* 0 */	 "l",
 /* 1 */	 "a,B",
 /* 2 */	 "$I,A,B",
 /* 3 */	 "B,a",
 /* 4 */	 "p",
 /* 5 */	 "$I,A,B",
 /* 6 */	 "B,A,p",
 /* 7 */	 "a",
 /* 8 */	 "PA,$I,B",
 /* 9 */	 "*",
 /* 10 */	 0,
 /* 11 */	 "$i,A,C",
 /* 12 */	 "B,A,C",
 /* 13 */	 "B,A,C",
 /* 14 */	 "A",
 /* 15 */	 "A,PC",
 /* 16 */	 "C",
 /* 17 */	 "M,C",
 /* 18 */	 "$i",
 /* 19 */	 "A,M",
};

static Opcode opcodes[64+64+1] = {
	/* J-type instructions */
 /* call	*/	ACALL,		0,
 /* jmpi	*/	AJMP,		0,

	/* I-type instructions */
				0,			0,
 /* ldbu	*/	AMOVBU,		1,
 /* addi	*/	AADD,		2,
 /* stb		*/	AMOVB,		3,
 /* br		*/	ABR,		4,
 /* ldb		*/	AMOVB,		1,
 /* cmpgei	*/	ACMPGE,		5,
				0,			0,
				0,			0,
 /* ldhu	*/	AMOVHU,		1,
 /* andi	*/	AAND,		2,
 /* sth		*/	AMOVH,		3,
 /* bge		*/	ABGE,		6,
 /* ldh		*/	AMOVH,		1,
 /* cmplti	*/	ACMPLT,		5,
				0,			0,
				0,			0,
 /* initda	*/	AINITDA,	7,
 /* ori		*/	AOR,		2,
 /* stw		*/	AMOVW,		3,
 /* blt		*/	ABLT,		6,
 /* ldw		*/	AMOVW,		1,
 /* cmpnei	*/	ACMPNE,		5,
				0,			0,
				0,			0,
 /* flushda	*/	AFLUSHDA,	7,
 /* xori	*/	AXOR,		2,
				0,			0,
 /* bne		*/	ABNE,		6,
				0,			0,
 /* cmpeqi	*/	ACMPEQ,		5,
				0,			0,
				0,			0,
 /* ldbuio	*/	AMOVBUIO,	1,
 /* muli	*/	AMUL,		2,
 /* stbio	*/	AMOVBIO,	3,
 /* beq		*/	ABEQ,		6,
 /* ldbio	*/	AMOVBIO,	1,
 /* cmpgeui	*/	ACMPGEU,	5,
				0,			0,
				0,			0,
 /* ldhuio	*/	AMOVHUIO,	1,
 /* andhi	*/	AAND,		2,
 /* sthio	*/	AMOVHIO,	3,
 /* bgeu	*/	ABGEU,		6,
 /* ldhio	*/	AMOVHIO,	1,
 /* cmpltui	*/	ACMPLTU,	5,
				0,			0,
 /* custom	*/	ACUSTOM,	9,
 /* initd	*/	AINITD,		7,
 /* orhi	*/	AOR,		2,
 /* stwio	*/	AMOVWIO,	3,
 /* bltu	*/	ABLTU,		6,
 /* ldwio	*/	AMOVWIO,	1,
 /* rdprs	*/	AADD,		8,
				0,			0,
				0,			0,
 /* flushd	*/	AFLUSHD,	7,
 /* xorhi	*/	AXOR,		2,
				0,			0,
				0,			0,
				0,			0,

	/* R-type instructions (op=0x3A, opx=bits 16..11) */
				0,			0,
 /* eret	*/	AERET,		10,
 /* roli	*/	AROL,		11,
 /* rol		*/	AROL,		12,
 /* flushp	*/	AFLUSHP,	10,
 /* ret		*/	ARET,		10,
 /* nor		*/	ANOR,		12,
 /* mulxuu	*/	AMULXUU,	12,
 /* cmpge	*/	ACMPGE,		13,
 /* bret	*/	ABRET,		10,
				0,			0,
 /* ror		*/	AROR,		12,
 /* flushi	*/	AFLUSHI,	14,
 /* jmp		*/	AJMP,		14,
 /* and		*/	AAND,		12,
				0,			0,
 /* cmplt	*/	ACMPLT,		13,
 				0,			0,
 /* slli	*/	ASLL,		11,
 /* sll		*/	ASLL,		12,
 /* wrprs	*/	AMOVW,		15,
				0,			0,
 /* or		*/	AOR,		12,
 /* mulxsu	*/	AMULXSU,	12,
 /* cmpne	*/	ACMPNE,		13,
				0,			0,
 /* srli	*/	ASRL,		11,
 /* srl		*/	ASRL,		12,
 /* nextpc	*/	ANEXTPC,	16,
 /* callr	*/	ACALL,		14,
 /* xor		*/	AXOR,		12,
 /* mulxss	*/	AMULXSS,	12,
 /* cmpeq	*/	ACMPEQ,		13,
				0,			0,
				0,			0,
				0,			0,
 /* divu	*/	ADIVU,		12,
 /* div		*/	ADIV,		12,
 /* rdctl	*/	AMOVW,		17,
 /* mul		*/	AMUL,		12,
 /* cmpgeu	*/	ACMPGEU,	13,
 /* initi	*/	AINIT,		14,
				0,			0,
				0,			0,
				0,			0,
 /* trap	*/	ATRAP,		18,
 /* wrctl	*/	AMOVW,		19,
				0,			0,
 /* cmpltu	*/	ACMPLTU,	13,
 /* add		*/	AADD,		12,
				0,			0,
				0,			0,
 /* break	*/	ABREAK,		18,
				0,			0,
 /* sync	*/	ASYNC,		10,
				0,			0,
				0,			0,
 /* sub		*/	ASUB,		12,
 /* srai	*/	ASRA,		11,
 /* sra		*/	ASRA,		12,
				0,			0,
				0,			0,
				0,			0,
				0,			0,

 /* ???		*/	AGOK,		10,
};

#pragma	varargck	argpos	bprint		2

static void
bprint(Instr *i, char *fmt, ...)
{
	va_list arg;

	va_start(arg, fmt);
	i->curr = vseprint(i->curr, i->end, fmt, arg);
	va_end(arg);
}

static void
format(Instr *i, char *mnemonic, char *f)
{
	int c;
	long imm;

	bprint(i, "%s", mnemonic);
	if(f == 0)
		return;
	bprint(i, "\t");
	for(; (c = *f); f++){
		switch(c){
		default:
			bprint(i, "%c", c);
			break;
		case 'A':
			bprint(i, "R%d", i->ra);
			break;
		case 'B':
			bprint(i, "R%d", i->rb);
			break;
		case 'C':
			bprint(i, "R%d", i->rc);
			break;
		case 'I':
			imm = i->imm16;
			switch(i->op){
			case 0x0C:	/* andi */
			case 0x14:	/* ori */
			case 0x1C:	/* xori */
				imm &= 0xFFFF;
				break;
			case 0x34:	/* orhi */
			case 0x3C:	/* xorhi */
			case 0x2C:	/* andhi */
				imm <<= 16;
				break;
			}
			if(imm < 0)
				bprint(i, "-%lux", -imm);
			else
				bprint(i, "%lux", imm);
			break;
		case 'i':
			imm = i->imm5;
			bprint(i, "%lux", imm);
			break;
		case 'l':
			imm = i->imm26<<2;
			bprint(i, "%lx", imm);
			break;
		case 'p':
			imm = i->addr + i->imm16 + 4;
			bprint(i, "%lx", imm);
			break;
		case 'a':
			bprint(i, "%x(R%d)", i->imm16, i->ra);
			break;
		case 'M':
			bprint(i, "CTL%d", i->imm5);
			break;
		case '*':
			imm = i->imm16;
			bprint(i, "%d,%c%d,%c%d,%c%d", imm&0xFF,
				i->w&(1<<16)? 'R' : 'C', i->ra,
				i->w&(1<<15)? 'R' : 'C', i->rb,
				i->w&(1<<14)? 'R' : 'C', i->ra);
			break;
		}
	}
}

static int
decode(Map *map, uvlong pc, Instr *i)
{
	ulong w;
	int op;

	if(get4(map, pc, &w) < 0) {
		werrstr("can't read instruction: %r");
		return -1;
	}
	i->w = w;
	op = (w&0x3F);
	if(op == 0x3A)
		op = 64 + ((w>>11)&0x3F);
	if(op > 64+64)
		op = 64+64;
	i->op = op;
	i->ra = (w>>27)&0x1F;
	i->rb = (w>>22)&0x1F;
	i->rc = (w>>17)&0x1F;
	i->imm26 = (w>>6)&0x3FFFFFF;
	i->imm16 = (w>>6)&0xFFFF;
	i->imm5 = (w>>6)&0x1F;
	i->addr = pc;
	i->map = map;
	return 1;
}

static int
niosdas(Map *map, uvlong pc, char modifier, char *buf, int n)
{
	Instr i;
	Opcode *o;

	USED(modifier);
	i.curr = buf;
	i.end = buf+n;
	if(decode(map, pc, &i) < 0)
		return -1;
	o = &opcodes[i.op];
	if(o->as == 0)
		format(&i, "???", 0);
	else
		format(&i, anames[o->as], fmts[o->fmt]);
	return 4;
}

static int
nioshexinst(Map *map, uvlong pc, char *buf, int n)
{
	Instr i;

	i.curr = buf;
	i.end = buf+n;
	if(decode(map, pc, &i) < 0)
		return -1;
	if(i.end-i.curr > 8)
		i.curr = _hexify(buf, i.w, 7);
	*i.curr = 0;
	return 4;
}

static int
niosinstlen(Map *map, uvlong pc)
{
	Instr i;

	if(decode(map, pc, &i) < 0)
		return -1;
	return 4;
}

/*
 *	Debugger interface
 */
Machdata niosmach =
{
	{0x3a, 0xa0, 0x3d, 0x00},		/* break point */
	4,			/* break point size */

	leswab,			/* short to local byte order */
	leswal,			/* long to local byte order */
	leswav,			/* long to local byte order */
	risctrace,		/* C traceback */
	riscframe,		/* Frame finder */
	0,				/* print exception */
	0,				/* breakpoint fixup */
	0,				/* single precision float printer */
	0,				/* double precisioin float printer */
	0,				/* following addresses */
	niosdas,		/* symbolic disassembly */
	nioshexinst,	/* hex disassembly */
	niosinstlen,	/* instruction size */
};
