%{
#include "a.h"
%}
%union
{
	Sym	*sym;
	long	lval;
	double	dval;
	char	sval[8];
	Gen	gen;
}
%left	'|'
%left	'^'
%left	'&'
%left	'<' '>'
%left	'+' '-'
%left	'*' '/' '%'
%token	<lval>	LMOVW LMOVB LMOVBU LADD LCMP LROL LDIV
%token	<lval>	LBEQ LBR LCALL LFLUSHD LFLUSHI LBREAK LBRET LCUSTOM LNEXTPC
%token	<lval>	LCONST LSP LSB LFP LPC LREG LR LC LCTL LPRS
%token	<lval>	LDATA LTEXT LWORD
%token	<sval>	LSCONST
%token	<dval>	LFCONST
%token	<sym>	LNAME LLAB LVAR
%type	<lval>	con expr pointer offset sreg creg
%type	<gen>	imr rreg xreg cregs ctlreg prsreg
%type	<gen>	addr name ireg oreg rel imm ximm fimm
%%
prog:
|	prog line

line:
	LLAB ':'
	{
		if($1->value != pc)
			yyerror("redeclaration of %s", $1->name);
		$1->value = pc;
	}
	line
|	LNAME ':'
	{
		$1->type = LLAB;
		$1->value = pc;
	}
	line
|	LNAME '=' expr ';'
	{
		$1->type = LVAR;
		$1->value = $3;
	}
|	LVAR '=' expr ';'
	{
		if($1->value != $3)
			yyerror("redeclaration of %s", $1->name);
		$1->value = $3;
	}
|	';'
|	inst ';'
|	error ';'

inst:
	LADD imm ',' prsreg ',' rreg
	{
		if($1 == AADD)
			outcode(ARDPRS, &$2, $4.reg, &$6);
		else
			yyerror("improper use of PRS reg");
	}
|	LADD imm ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LADD rreg ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LADD imm ',' sreg ',' rreg
	{
		outcode($1, &$2, $4, &$6);
	}
|	LADD rreg ',' sreg ',' rreg
	{
		outcode($1, &$2, $4, &$6);
	}

|	LBEQ rreg ',' sreg ',' rel
	{
		outcode($1, &$2, $4, &$6);
	}

|	LBR	rel
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LBREAK imm
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LBRET
	{
		outcode($1, &nullgen, NREG, &nullgen);
	}

|	LCALL name
	{
		outcode($1, &nullgen, NREG, &$2);
	}
|	LCALL rel
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LCMP imr ',' sreg ',' rreg	
	{
		outcode($1, &$2, $4, &$6);
	}

|	LCUSTOM imm ',' cregs
	{
		outcode($1, &$2, NREG, &$4);
	}

|	LDIV rreg ',' sreg ',' rreg
	{
		outcode($1, &$2, $4, &$6);
	}

|	LFLUSHD oreg
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LFLUSHI ireg
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LNEXTPC rreg
	{
		outcode($1, &nullgen, NREG, &$2);
	}

|	LROL imr ',' sreg ',' rreg
	{
		outcode($1, &$2, $4, &$6);
	}
|	LROL imr ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}

|	LMOVB addr ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVB rreg ',' addr
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVB rreg ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}

|	LMOVBU rreg ',' addr
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVBU rreg ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}

|	LMOVW imm ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW ximm ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW rreg ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW addr ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW rreg ',' addr
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW rreg ',' xreg
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LMOVW xreg ',' rreg
	{
		outcode($1, &$2, NREG, &$4);
	}

|	LTEXT name ',' imm
	{
		outcode($1, &$2, NREG, &$4);
	}
|	LTEXT name ',' con ',' imm
	{
		outcode($1, &$2, $4, &$6);
	}

|	LDATA name '/' con ',' imm
	{
		outcode($1, &$2, $4, &$6);
	}
|	LDATA name '/' con ',' ximm
	{
		outcode($1, &$2, $4, &$6);
	}
|	LDATA name '/' con ',' fimm
	{
		outcode($1, &$2, $4, &$6);
	}

|	LWORD imm
	{
		outcode($1, &nullgen, NREG, &$2);
	}

rel:
	con '(' LPC ')'
	{
		$$ = nullgen;
		$$.type = D_BRANCH;
		$$.offset = $1 + pc;
	}
|	LNAME offset
	{
		$$ = nullgen;
		if(pass == 2)
			yyerror("undefined label: %s", $1->name);
		$$.type = D_BRANCH;
		$$.sym = $1;
		$$.offset = $2;
	}
|	LLAB offset
	{
		$$ = nullgen;
		$$.type = D_BRANCH;
		$$.sym = $1;
		$$.offset = $1->value + $2;
	}

addr:
	oreg
|	name

oreg:
	'(' sreg ')'
	{
		$$ = nullgen;
		$$.type = D_OREG;
		$$.reg = $2;
		$$.offset = 0;
	}
|	con '(' sreg ')'
	{
		$$ = nullgen;
		$$.type = D_OREG;
		$$.reg = $3;
		$$.offset = $1;
	}

ireg:
	oreg
	{
		if($$.offset != 0)
			yyerror("register offset must be zero");
	}

name:
	con '(' pointer ')'
	{
		$$ = nullgen;
		$$.type = D_OREG;
		$$.name = $3;
		$$.sym = S;
		$$.offset = $1;
	}
|	LNAME offset '(' pointer ')'
	{
		$$ = nullgen;
		$$.type = D_OREG;
		$$.name = $4;
		$$.sym = $1;
		$$.offset = $2;
	}
|	LNAME '<' '>' offset '(' LSB ')'
	{
		$$ = nullgen;
		$$.type = D_OREG;
		$$.name = D_STATIC;
		$$.sym = $1;
		$$.offset = $4;
	}

offset:
	{
		$$ = 0;
	}
|	'+' con
	{
		$$ = $2;
	}
|	'-' con
	{
		$$ = -$2;
	}

pointer:
	LSB
|	LSP
|	LFP

imr:
	imm
|	rreg

rreg:
	sreg
	{
		$$ = nullgen;
		$$.type = D_REG;
		$$.reg = $1;
	}

xreg:
	ctlreg
|	prsreg

sreg:
	LREG
|	LR '(' expr ')'
	{
		if($3 < 0 || $3 >= NREG)
			yyerror("register value out of range");
		$$ = $3;
	}

cregs:
	creg ',' creg ',' creg
	{
		$$ = nullgen;
		$$.type = D_CONST;
		$$.offset = ($1<<16) | ($3<<8) | $5;
	}

creg:
	sreg
|	LC '(' expr ')'
	{
		if($3 < 0 || $3 >= NREG)
			yyerror("custom register value out of range");
		$$ = NREG+$3;
	}

ctlreg:
	LCTL '(' expr ')'
	{
		if($3 < 0 || $3 >= NREG)
			yyerror("custom register value out of range");
		$$ = nullgen;
		$$.type = D_CTLREG;
		$$.reg = $3;
	}

prsreg:
	LPRS '(' expr ')'
	{
		if($3 < 0 || $3 >= NREG)
			yyerror("PRS register value out of range");
		$$ = nullgen;
		$$.type = D_PRSREG;
		$$.reg = $3;
	}

ximm:
	'$' addr
	{
		$$ = $2;
		$$.type = D_CONST;
	}
|	'$' LSCONST
	{
		$$ = nullgen;
		$$.type = D_SCONST;
		memcpy($$.sval, $2, sizeof($$.sval));
	}

fimm:
	'$' LFCONST
	{
		$$ = nullgen;
		$$.type = D_FCONST;
		$$.dval = $2;
	}
|	'$' '-' LFCONST
	{
		$$ = nullgen;
		$$.type = D_FCONST;
		$$.dval = -$3;
	}

imm:
	'$' con
	{
		$$ = nullgen;
		$$.type = D_CONST;
		$$.offset = $2;
	}

con:
	LCONST
|	LVAR
	{
		$$ = $1->value;
	}
|	'-' con
	{
		$$ = -$2;
	}
|	'+' con
	{
		$$ = $2;
	}
|	'~' con
	{
		$$ = ~$2;
	}
|	'(' expr ')'
	{
		$$ = $2;
	}

expr:
	con
|	expr '+' expr
	{
		$$ = $1 + $3;
	}
|	expr '-' expr
	{
		$$ = $1 - $3;
	}
|	expr '*' expr
	{
		$$ = $1 * $3;
	}
|	expr '/' expr
	{
		$$ = $1 / $3;
	}
|	expr '%' expr
	{
		$$ = $1 % $3;
	}
|	expr '<' '<' expr
	{
		$$ = $1 << $4;
	}
|	expr '>' '>' expr
	{
		$$ = $1 >> $4;
	}
|	expr '&' expr
	{
		$$ = $1 & $3;
	}
|	expr '^' expr
	{
		$$ = $1 ^ $3;
	}
|	expr '|' expr
	{
		$$ = $1 | $3;
	}
